;;; funnel.el - Major mode for editing Funnel code.
;;; $Id: funnel-mode.el,v 1.4 2001/03/28 07:22:45 schinz Exp $

;; TODO:
;; Font-lock : - take records into account

(require 'easymenu)
(require 'inferior-funnel-mode)

(defconst funnel-mode-version "0.2.1")
(defconst funnel-bug-e-mail "Michel.Schinz@epfl.ch")
(defconst funnel-web-url "http://lampwww.epfl.ch/funnel/")

(defconst funnel-xemacsp (string-match "XEmacs" emacs-version)
  "True iff we are running in XEmacs")

;; Customization

(defgroup funnel
  nil
  "Mode to edit Funnel code."
  :group 'languages)

(defcustom funnel-indent-step 2
  "Indentation step."
  :type 'integer
  :group 'funnel)

(defcustom funnel-file-suffix "fnl?"
  "Suffix for Funnel source code files."
  :type 'string
  :group 'funnel)

(defcustom funnel-mode-hook nil
  "Hook run when Funnel mode starts."
  :type 'hook
  :group 'funnel)

;; Key mappings

(defvar funnel-mode-map nil
  "Keymap used in Funnel mode.")

(unless funnel-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\t" 'funnel-indent-line)
    (define-key map "\M-\t" 'funnel-outdent-line)
    (define-key map [backspace] 'backward-delete-char-untabify)
    (define-key map "}" 'funnel-electric-brace)

    (define-key map "\C-c\C-l" 'funnel-load-file)
    (define-key map "\C-c\C-r" 'funnel-eval-region)
    (define-key map "\C-c\C-b" 'funnel-eval-buffer)
    (define-key map "\C-c\C-z" 'funnel-switch-to-interpreter)

    (easy-menu-define funnel-menu-bar map "Funnel menu"
		      '("Funnel"
			["Run interpreter..."
			 run-funnel (not (funnel-interpreter-runningp))]
			["Reset interpreter"
			 funnel-reset-interpreter (funnel-interpreter-runningp)]
			["Quit interpreter"
			 funnel-quit-interpreter (funnel-interpreter-runningp)]
			["Load file in interpreter..."
			 funnel-load-file (funnel-interpreter-runningp)]
			["Switch to interpreter"
			 funnel-switch-to-interpreter (funnel-interpreter-runningp)]
			["Evaluate region"
			 funnel-eval-region (and (funnel-interpreter-runningp)
						 (funnel-region-definedp))]
			["Evaluate buffer"
			 funnel-eval-buffer (funnel-interpreter-runningp)]
			"---"
			["Browse Funnel Web site..."
			 funnel-browse-web-site t]
			["Customize..."
			 funnel-customize t]
			["Report bug..."
			 funnel-report-bug t]))
    (setq funnel-mode-map map)))

;; Syntax table

(defvar funnel-mode-syntax-table nil
  "Syntax table in use in Funnel mode buffers.")

(unless funnel-mode-syntax-table
  (let ((table (make-syntax-table)))
    ;; special characters
    (modify-syntax-entry ?\! "_" table)
    (modify-syntax-entry ?\@ "_" table)
    (modify-syntax-entry ?\$ "_" table)
    (modify-syntax-entry ?\% "_" table)
    (modify-syntax-entry ?\; "_" table)
    (modify-syntax-entry ?\^ "_" table)
    (modify-syntax-entry ?\& "_" table)
    (modify-syntax-entry ?\- "_" table)
    (modify-syntax-entry ?\_ "_" table)
    (modify-syntax-entry ?\+ "_" table)
    (modify-syntax-entry ?\= "_" table)
    (modify-syntax-entry ?\~ "_" table)
    (modify-syntax-entry ?\< "_" table)
    (modify-syntax-entry ?\> "_" table)
    (modify-syntax-entry ?\| "_" table)
    (modify-syntax-entry ?\? "_" table)
    (modify-syntax-entry ?\: "_" table)
    (modify-syntax-entry ?\ "_" table)
    (modify-syntax-entry ?\ "_" table)

    ;; punctuation characters
    (modify-syntax-entry ?\. "." table)

    ;; comments
    (if funnel-xemacsp
	(modify-syntax-entry ?\/  ". 1456" table)
      (modify-syntax-entry ?\/  ". 124b" table))
    (modify-syntax-entry ?\*  "_ 23"   table)
    (modify-syntax-entry ?\n "> b" table)
    (modify-syntax-entry ?\r "> b" table)
    (setq funnel-mode-syntax-table table)))

;; Indentation

(defvar funnel-cached-pps-to nil)
(make-variable-buffer-local 'funnel-cached-pps-to)
(defvar funnel-cached-pps-state nil)
(make-variable-buffer-local 'funnel-cached-pps-state)

(defun funnel-parse-partial-sexp ()
  "Equivalent to `parse-partial-sexp' from buffer start to point.
Tries to perform some cacheing to avoid rescanning the whole buffer often."
  (let ((from (point-min)) (to (point)))
    (unless (and funnel-cached-pps-to (= funnel-cached-pps-to to))
      (let* ((can-reuse-statep (and funnel-cached-pps-to (> to funnel-cached-pps-to)))
	     (real-from (if can-reuse-statep funnel-cached-pps-to from))
	     (real-state (if can-reuse-statep funnel-cached-pps-state nil)))
	(setq funnel-cached-pps-state
	      (parse-partial-sexp real-from to nil nil real-state))
	(setq funnel-cached-pps-to to)))
    funnel-cached-pps-state))

(defun funnel-clear-pps-cache (start end)
  "Clear cache used by `funnel-parse-partial-sexp'.
This function should be called whenever the buffer is changed. That is, it
should be part of `before-change-functions'."
  (when (and funnel-cached-pps-to
	     (< start funnel-cached-pps-to))
    (setq funnel-cached-pps-to nil)))

(defun funnel-last-definition-pos ()
  "Position of last (easily) identifiable top-level definition before point.
This can be used as a FROM argument for `parse-partial-sexp'"
  (point-min))				;[...] refine

(defun funnel-in-indented-blockp ()
  "True iff the point is in a block where identation matters."
  (let ((block-start (nth 1 (funnel-parse-partial-sexp))))
    (or (not block-start)
	(char-equal (char-after block-start) ?{))))

(defun funnel-at-indented-block-startp ()
  "True iff the point is before the first expression in an indented block.
See also `funnel-in-indented-blockp'."
  (save-excursion
    (forward-comment (- (buffer-size)))
    (char-equal (preceding-char) ?{)))

(defun funnel-in-indentationp ()
  "True iff the point is in the indentation of the current line."
  (<= (current-column) (current-indentation)))

(defun funnel-in-commentp ()
  "True iff the point is inside a comment."
  (nth 4 (funnel-parse-partial-sexp)))

(defun funnel-in-literalp ()
  "True iff the point is inside a string literal.

Works only if outside a comment (i.e. `funnel-in-commentp' must be
false for that function to work correctly)!"
  ;; Since strings cannot span across lines in Funnel, we do not need
  ;; to use the full power of `funnel-parse-partial-sexp'.
  (save-excursion
    (save-match-data
      (let ((pt (point)))
	(beginning-of-line)
	;; Move past all well-terminated strings
	(while (search-forward-regexp "\"\\([^\\\\\"]\\|\\\\\"\\)*\"" pt t))
	(search-forward-regexp "\"\\([^\\\\\"]\\|\\\\\"\\)*" pt t)))))

(defun funnel-empty-or-comment-line-p ()
  "True iff the current line is empty or made of comments only.
Does *not* work inside of a multi-line comment."
  (save-excursion
    (let ((eol-pos (progn (end-of-line) (point))))
      (beginning-of-line)
      (forward-comment (buffer-size))
      (>= (point) eol-pos))))

(defun funnel-current-line-indentation ()
  "Return indentation of current line."
  (save-excursion
    (end-of-line)
    (while (or (not (bolp))
	       (funnel-empty-or-comment-line-p))
      (while (funnel-empty-or-comment-line-p)
	(forward-line -1)
	(end-of-line))
      (let ((bol-pos (save-excursion (beginning-of-line) (point))))
	(skip-syntax-backward "^)" bol-pos))
      (unless (bolp) (backward-sexp)))
    (current-indentation)))

(defun funnel-indent-line-to (column)
  "Indent current line to COLUMN and perhaps move point.
The point is moved iff it is currently in the indentation, in which
case it is brought to the end of that indentation. Otherwise it does
not move."
  (if (funnel-in-indentationp)
      (indent-line-to column)
    (save-excursion (indent-line-to column))))

(defun funnel-indentation ()
  "Return the suggested indentation for the current line."
  (save-excursion
    (let* ((state (funnel-parse-partial-sexp))
	   (block-start (nth 1 state)))
      (if block-start
	  (progn
	    (goto-char (1+ block-start))
	    (let ((block-start-eol (save-excursion (end-of-line) (point)))
		  (block-start-column (current-column)))
	      (while (progn (skip-syntax-forward " ")
			    (funnel-forward-bindings)
			    (skip-syntax-forward " ")
			    (char-equal (following-char) ?\|))
		(forward-char))
	      (if (or (= (point) (1+ block-start))
		      (and (= (point) block-start-eol)
			   (char-equal (char-before) ?\|)))
		  (+ (funnel-current-line-indentation)
		     funnel-indent-step)
		block-start-column)))
	0))))

(defun funnel-indent-rigidly ()
  "Indent like previous line or one step further right.
When at the beginning of the line, indent like the previous non-blank line.
Otherwise indent one step further right."
  (let ((new-indentation (if (bolp)
			     (max funnel-indent-step (funnel-indentation))
			   (+ (current-indentation) funnel-indent-step))))
    (funnel-indent-line-to new-indentation)))

(defun funnel-outdent-rigidly ()
  "Indent just right of previous line or one step further left.
When at the beginning of the line, indent one step left of the previous
non-blank line. Otherwise, indent one step further left."
  (let ((new-indentation (max 0 (if (bolp)
				    (- (funnel-indentation) funnel-indent-step)
				  (- (current-indentation) funnel-indent-step)))))
    (funnel-indent-line-to new-indentation)))

(defun funnel-indent-line ()
  "Indent current line as smartly as possible."
  (interactive)
  (if (funnel-in-commentp)
      (insert ?\t)
    (funnel-indent-rigidly)))

(defun funnel-outdent-line ()
  "Outdent current line."
  (interactive)
  (funnel-outdent-rigidly))

(defun funnel-electric-brace (arg)
  "Insert a brace.

The current line is also reindented provided that:
* no numeric argument is supplied,
* the brace is on a line by itself (i.e. surrounded by spaces only), and
* the brace is not inside a comment or a string literal."
  (interactive "*P")
  (unless (or arg
	      (not (funnel-in-indentationp))
	      (not (looking-at "[ \t]*$"))
	      (funnel-in-commentp)
	      (funnel-in-literalp))
    (funnel-indent-line-to (max 0 (- (funnel-indentation) funnel-indent-step))))
  (self-insert-command (prefix-numeric-value arg)))

;; Font lock

(defconst funnel-type-chars "-a-zA-Z.> \t\n"
  "Characters that can appear in a type expression, braces excluded.")

(defconst funnel-special-chars "!@%;\\^&*+=~<>|?:\\-"
  "Characters which are considered special by the Funnel syntax.")

(defconst funnel-special-chars-list
  (delete ?\\ (string-to-list funnel-special-chars))
  "Characters which are considered special by the Funnel syntax, as a list.")

(defconst funnel-letter-chars "a-zA-Z$_"
  "Characters which can begin non-special Funnel identifiers.")

(defconst funnel-letter-or-digit-chars (concat "0-9" funnel-letter-chars)
  "Characters which can form non-special Funnel identifiers.")

(defun funnel-special-char-p (c)
  (member c funnel-special-chars-list))

(defconst funnel-identifier-begin-re (concat "["
					     funnel-letter-or-digit-chars
					     funnel-special-chars
					     "]"))

(defun funnel-at-identifier-begin-p ()
  (looking-at funnel-identifier-begin-re))

(defun funnel-forward-type ()
  "Move forward across one Funnel type expression.

For this function to work, the point has to be at the beginning of a
type expression."
  (while (progn (skip-chars-forward funnel-type-chars)
		(and (not (eobp))
		     (or (eq (following-char) ?\()
			 (eq (following-char) ?\[))))
    (forward-sexp)))

(defun funnel-forward-identifier ()
  "Move forward across the Funnel identifier starting at point."
  (if (funnel-special-char-p (following-char))
      (skip-chars-forward funnel-special-chars)
    (skip-chars-forward funnel-letter-or-digit-chars))
  (skip-chars-forward "'"))

(defun funnel-forward-qualified-identifier ()
  "Move forward across the Funnel (maybe qualified) identifier starting at point."
  (while (progn (funnel-forward-identifier)
		(and (not (eobp))
		     (and (eq (following-char) ?\.)
			  (progn (forward-char) t))))))

(defun funnel-forward-binding ()
  "Move forward across one binding."
  (funnel-forward-identifier)
  (when (looking-at "\\s *:\\s *")
    (goto-char (match-end 0))
    (funnel-forward-type)))

(defun funnel-forward-bindings ()
  "Move forward across several bindings separated by commas."
  (while (progn (funnel-forward-binding)
		(skip-syntax-forward " ")
		(char-equal (following-char) ?\,))
    (forward-char)
    (skip-syntax-forward " ")))

(defun funnel-delimit-type ()
  "Delimit the type following the point.

After executing this function, the whole match will refer to the
type."
  (save-excursion
    (let ((type-beg (point-marker)))
      (funnel-forward-type)
      (set-match-data (list type-beg (point-marker))))))

(defun funnel-delimit-qualified-identifier ()
  "Delimit the qualified identifier following the point.

After executing this function, the whole match will refer to the
identifier."
  (save-excursion
    (skip-syntax-forward " ")
    (let ((beg (point-marker)))
      (funnel-forward-qualified-identifier)
      (set-match-data (list beg (point-marker))))))

(defun funnel-delimit-type-parameters ()
  "Delimit the type parameters following the point.

After executing this function, the whole match will refer to the type
parameters."
  (save-excursion
    (let ((beg (point-marker)))
      (while (progn (skip-syntax-forward " ")
		    (and (eq (following-char) ?\[) (not (eobp))))
	(forward-sexp))
      (set-match-data (list beg (point-marker))))))

(defconst funnel-binding-regexp "\\s *\\(\\sw+\\)"
  "Regexp matching the beginning of a binding.")

(defun funnel-delimit-binding ()
  "Delimit the binding following the point.

After executing this function, submatch 1 will refer to the identifier
bound, and submatch 2 to its type, if any."
  (save-excursion
    (funnel-delimit-qualified-identifier)
    (let ((id-beg (match-beginning 0)) (id-end (match-end 0))
	  type-beg type-end)
      (goto-char id-end)
      (skip-syntax-forward " ")
      (when (eq (following-char) ?\:)
	(forward-char)
	(skip-syntax-forward " ")
	(setq type-beg (point-marker))
	(funnel-forward-type)
	(setq type-end (point-marker)))
      (set-match-data (list id-beg (or type-end id-end)
			    id-beg id-end
			    type-beg type-end)))))

(defun funnel-match-binding-and-skip (limit)
  "Internal function used for highlighting."
  (save-restriction
    (narrow-to-region (point-min) limit)
    (condition-case nil
	(when (looking-at funnel-binding-regexp)
	  (funnel-delimit-binding)
	  (goto-char (match-end 0))
	  t)
      (error nil))))

(defun funnel-match-pattern-and-skip (limit)
  "Internal function used for highlighting."
  (save-restriction
    (narrow-to-region (point-min) limit)
    (condition-case nil
	(progn
	  (skip-chars-forward "() \t\n,")
	  (when (and (looking-at funnel-binding-regexp)
		     (not (eq (following-char) ?\=)))
	    (funnel-delimit-binding)
	    (goto-char (match-end 0))
	    t))
      (error nil))))

(defvar funnel-font-lock-face nil
  "The face to use for the next fontification (internal use).")

(defun funnel-match-rassoc-and-skip (limit)
  "Internal function used for highlighting."
  (save-restriction
    (narrow-to-region (point-min) limit)
    (condition-case nil
	(progn
	  (skip-syntax-forward " ")
	  (cond ((or (eobp)
		     (and (eq (following-char) ?\=)
			    (not (and (char-after (1+ (point)))
				      (funnel-special-char-p (char-after
							      (1+ (point))))))))
		 nil)

		((eq (following-char) ?\))
		 ;; Formals are terminated, the identifier follows,
		 ;; maybe with a type.
		 (forward-char)
		 (when (looking-at funnel-binding-regexp)
		   (funnel-delimit-qualified-identifier)
		   (goto-char (match-end 0))
		   (let ((ident-beg (match-beginning 0)) (ident-end (match-end 0)))
		     (if (looking-at "\\s *:\\s *")
			 (progn
			   (goto-char (match-end 0))
			   (let ((type-beg (point-marker)))
			     (funnel-forward-type)
			     (let ((type-end (point-marker)))
			       (set-match-data (list ident-beg type-end
						     ident-beg ident-end
						     type-beg type-end)))))
		       (set-match-data (list ident-beg ident-end
					     ident-beg ident-end))))
		   (setq funnel-font-lock-face font-lock-function-name-face)
		   t))

		((looking-at (concat "\\s *\\(\\<def\\>\\)?" funnel-binding-regexp))
		 ;; One part of a formal
		 (when (match-end 1) (goto-char (match-end 1)))
		 (funnel-delimit-binding)
		 (goto-char (match-end 0))
		 (skip-chars-forward " \t\n,")
		 (setq funnel-font-lock-face font-lock-variable-name-face)
		 t)))
      (error nil))))

(defun funnel-match-lassoc-and-skip (limit)
  "Internal function used for highlighting."
  (save-restriction
    (narrow-to-region (point-min) limit)
    (condition-case nil
	(progn
	  (skip-syntax-forward " ")
	  (let ((next-char (following-char)))
	    (cond ((or (eobp)
		       (and (eq (following-char) ?\=)
			    (not (and (char-after (1+ (point)))
				      (funnel-special-char-p (char-after
							      (1+ (point))))))))
		   nil)

		  ((and (eq next-char ?\&)
			(not (funnel-special-char-p (char-after (1+ (point))))))
		   ;; A join pattern
		   (forward-char)
		   (skip-syntax-forward "-"))

		  ((and (eq next-char ?\:)
			(not (funnel-special-char-p (char-after (1+ (point))))))
		   ;; The result type
		   (skip-chars-forward " \t\n:")
		   (let ((type-beg (point-marker)))
		     (funnel-forward-type)
		     (let ((type-end (point-marker)))
		       (set-match-data (list type-beg type-end
					     type-beg type-end))))
		   (setq funnel-font-lock-face font-lock-type-face)
		   t)

		  ((or (eq next-char ?\() (eq next-char ?\,))
		   ;; A formal
		   (skip-chars-forward " \t\n(,")
		   (when (looking-at "\\s *\\<def\\>") (goto-char (match-end 0)))
		   (when (looking-at funnel-binding-regexp)
		     (funnel-delimit-binding)
		     (goto-char (match-end 0))
		     (skip-chars-forward " \t\n)")
		     (setq funnel-font-lock-face font-lock-variable-name-face)
		     t))

		  ((funnel-at-identifier-begin-p)
		   ;; An identifier (possibly qualified)
		   (funnel-delimit-qualified-identifier)
		   (goto-char (match-end 0))
		   (let ((id-beg (match-beginning 0)) (id-end (match-end 0)))
		     (funnel-delimit-type-parameters)
		     (goto-char (match-end 0))
		     (let ((param-beg (match-beginning 0)) (param-end (match-end 0)))
		       (set-match-data (list id-beg param-end
					     id-beg id-end
					     param-beg param-end))))
		   (setq funnel-font-lock-face font-lock-function-name-face)
		   t))))
      (error nil))))

(defconst funnel-keywords
  (regexp-opt '("|" "&" "," ":=" "<:" "<<" "=" ">>" "_"
		"abstract" "case" "class" "def" "final" "handle"
		"instance" "let" "module" "new" "private" "static"
		"throw " "try" "type" "val" "var" "with")
	      t))

(defconst funnel-constants
  (regexp-opt '("true" "false" "null" "this" "super") t))

(defconst funnel-definitions
  (regexp-opt '("var" "val" "def" "let") t))

(defvar funnel-font-lock-keywords
  `(;; keywords
    (,(concat "\\<" funnel-keywords "\\>") 0 font-lock-keyword-face nil)

    ;; constants
    (,(concat "\\<" funnel-constants "\\>")
     0 ,(if (boundp 'font-lock-constant-face)
	    'font-lock-constant-face
	  'font-lock-keyword-face)
     nil nil)

    ;; variables and lazy records
    ("\\<\\(var\\|let\\)\\>"
     (funnel-match-binding-and-skip (goto-char (match-end 0))
				    nil
				    (1 font-lock-variable-name-face nil)
				    (2 font-lock-type-face nil t)))

    ;; values
    ("\\<val\\>"
     (funnel-match-pattern-and-skip (goto-char (match-end 0))
				    nil
				    (1 font-lock-variable-name-face nil)
				    (2 font-lock-type-face nil t)))

    ;; right-associative definitions
    ("\\(^\\|[^(, \t\n]\\)\\s *\\<def\\>\\s +("
     (funnel-match-rassoc-and-skip (goto-char (match-end 0))
				   (goto-char (match-end 0))
				   (1 funnel-font-lock-face nil)
				   (2 font-lock-type-face nil t)))

    ;; left-associative (i.e. "standard") definitions
    ("\\(^\\|[^(, \t\n]\\)\\s *\\<def\\>\\s +[^(]"
     (funnel-match-lassoc-and-skip (goto-char (1- (match-end 0)))
				   nil
				   (1 funnel-font-lock-face nil t)
				   (2 font-lock-type-face nil t)))

    ;; closure definitions
    ;; ...
    ))

;; Misc

(defun funnel-region-definedp ()
  "True iff the region is currently defined."
  (condition-case nil
      (progn (region-beginning) t)
    (error nil)))

(defun funnel-browse-web-site ()
  "Browse the Funnel home-page"
  (interactive)
  (require 'browse-url)
  (browse-url funnel-web-url))

;; Bug reporting

(defun funnel-report-bug ()
  "Report a bug to the author of the Funnel mode via e-mail.
The package used to edit and send the e-mail is the one selected
through `mail-user-agent'."
  (interactive)
  (require 'reporter)
  (let ((reporter-prompt-for-summary-p t))
    (reporter-submit-bug-report
     funnel-bug-e-mail
     (concat "Funnel mode v" funnel-mode-version)
     '(funnel-indent-step
       funnel-file-suffix
       funnel-mode-hook))))

;; Customization

(defun funnel-customize ()
  "Customize Funnel mode using the Customize package."
  (interactive)
  (customize-group 'funnel))

;; Mode

(defun funnel-mode ()
  "Major mode for editing Funnel code.

When started, run `funnel-mode-hook'.

\\{funnel-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'funnel-mode)
  (setq mode-name "Funnel")

  ;; Font lock
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(funnel-font-lock-keywords
			     nil nil nil))

  ;; Syntax table
  (set-syntax-table funnel-mode-syntax-table)

  ;; Paragraph separation
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^\\s *$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)

  ;; Comment handling
  (make-local-variable 'comment-start)
  (setq comment-start "// ")
  (make-local-variable 'comment-end)
  (setq comment-end "")
  (make-local-variable 'comment-multi-line)
  (setq comment-multi-line nil)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "// *")

  ;; Misc
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)

  (make-local-hook 'before-change-functions)
  (add-hook 'before-change-functions
	    (function funnel-clear-pps-cache) nil t)

  ;; Keymap & menus
  (use-local-map funnel-mode-map)
  (easy-menu-add funnel-menu-bar)

  (run-hooks 'funnel-mode-hooks))

;; Automatic mode selection for funnel files, if not already done.
(let ((suffix-regexp (concat "\\." funnel-file-suffix "\\'")))
  (unless (assoc suffix-regexp auto-mode-alist)
    (setq auto-mode-alist (cons (cons suffix-regexp 'funnel-mode)
				auto-mode-alist))))

;; Make sure that Funnel files are not ignored by completion or by
;; dired (this is typically the case if the standard ".fn" extension
;; is used).
(let ((funnel-extension (concat "." funnel-file-suffix)))
  (setq completion-ignored-extensions
	(delete funnel-extension completion-ignored-extensions)))

(provide 'funnel-mode)
